import numpy as np
from multiprocessing import Process, Queue

from pyarchi.utils import create_logger

logger = create_logger("utils")


def run_function(queue, func, factors, data_f, to_disable=[], **kwargs):
    """
    Run each interaction of the function and  returns the results ordered on a dictionary

    Parameters
    ----------
    queue
    func
    factors

    Returns
    -------

    """

    kwargs_optim = kwargs.copy()

    kwargs_optim["debug"] = 0
    kwargs_optim["show_results"] = 0
    kwargs_optim["plot_realtime"] = 0

    kwargs_optim["report_pictures"] = 0
    kwargs_optim["show_results"] = 0
    kwargs_optim["export_text"] = 0
    kwargs_optim["export_fit"] = 0
    kwargs_optim["low_memory"] = 1
    kwargs_optim["uncertainties"] = 0

    results_dict = {}
    data_f.load_parameters(factors[0], **kwargs_optim)

    for star_ind in to_disable:
        data_f.disable_star(star_ind)

    for index, fac in enumerate(factors):

        if index != 0:
            data_f.load_parameters(fac, **kwargs_optim)

        data_fits = func(DataFits=data_f, factor=fac, **kwargs_optim)

        star_results = {}
        for ind, star in enumerate(data_fits.stars):
            star_results[ind] = star.calculate_cdpp(data_fits.mjd_time)[0]

            if star.out_bound:
                star_results[ind] = float("nan")

        results_dict[fac] = star_results

    queue.put(results_dict)
    return


def optimizer(value_range, max_process, func, data_f, file_path, to_disable=[], **kwargs):
    """
    Decides which factors will be used in each process that it spawns. After the processes are done, extracts the
    resulting data from the Queue and parses it, in order to find the noise values and mask factors. Writes to a .txt
    file the resulting values for each factor. Take note that the values are not guaranteed to be in order, since they
    are saved in the order "imposed" by the processes.
    Parameters
    ----------
    value_range:
        Lower and upper limit of the values to be tested
    max_process:
        Maximum number of processes that can be launched during this routine
    func:
        function that launches the photometric process
    data_f:
        :class:`pyarchi.main.initial_loads.Data`  object with all the stars information inside
    file_path:
        Path in which run time information shall be stored
    kwargs

    Returns
    -------
    min_cvs:
        list with the minimum noise found

    optimized_dict:
        Dictionary in which the keys are the star numbers and the values are the optimal factors
    """

    comms_queue = Queue()
    process_to_spawn = (
        max_process if value_range.shape[0] >= max_process else value_range.shape[0]
    )

    logger.info(
        "Optimizer going to spawn {} processes, for values: {}".format(
            process_to_spawn, value_range
        )
    )
    # Divide into equal lists the factors to be used

    splitted_values = np.array_split(value_range, process_to_spawn)

    # elements not inside the sublists

    logger.debug(splitted_values)

    for k in range(process_to_spawn):
        p = Process(
            target=run_function,
            args=(comms_queue, func, splitted_values[k], data_f, to_disable),
            kwargs=kwargs,
        )
        p.start()

    factors = []
    cvs = []

    for ii in range(process_to_spawn):

        data = comms_queue.get()
        if (
            ii == 0
        ):  # For the first run, create list with list for each star, in which the noise values will be stored
            cvs = [[] for _ in range(len(data[list(data.keys())[0]]))]

        for factor, stars_dict in data.items():
            factors.append(factor)

            for key, CV in stars_dict.items():
                cvs[key].append(CV)

    factors = np.array(factors)
    optimized_dict = {}

    min_cvs = {}
    for index, star_list in enumerate(cvs):
        min_cv = np.nanmin(star_list)
        if np.isnan(min_cv):
            optimized_dict[str(index)] = 1
            min_cvs[index] = 2e7
            continue

        min_cvs[index] = min_cv
        min_index = np.where(star_list == min_cv)

        optimal_factor = factors[min_index]

        optimized_dict[str(index)] = int(optimal_factor[0])

    with open(file_path, mode="a") as file:

        for index, fac in enumerate(factors):
            cdpps = ""
            for star_list in cvs:
                cdpps += str(star_list[index]) + " "

            file.write("{} \t {} \n".format(fac, cdpps))

    return min_cvs, optimized_dict



from astropy.io import fits
import os

from pyarchi.utils import create_logger
from pyarchi.utils import path_finder

logger = create_logger("util")


def create_fits(master_folder, data_fits, **kwargs):
    """
    Export the results to a fits file, containing the light curves and a correspondence of star to Cv and
     respective radius factor. 



    Parameters
    ----------
        master_folder:
            Path in which the data shall be stored
        data_fits
            :class:`~pyarchi.data_objects.Data.Data` object.
        kwargs


    Notes
    -----

        Data stored in the header unit of the file :
            Keyword      data
            
            method        type of mask used
            detect         tracking method 
            initial        initial detection method 
            grid           size of the background grid
            CDPP_TYPE       CDPP algorithm in use
        
       In the data unit of the file, we have each star, with the corresponding
        time, rotation angle, flux values and uncertainties
    """

    # TODO: store more info -> improve organization

    logger.info("Extracting data to FITS file")
    hdus = []
    default_path = path_finder(mode="default", **kwargs)
    try:
        hdulist = fits.open(default_path)
    except IOError:
        logger.fatal("File does not exist")
        return -1
    else:
        with hdulist:
            roll_ang = hdulist[1].data["ROLL_ANGLE"]
            mjd_time = hdulist[1].data["MJD_TIME"]

    col1 = fits.Column(name="MJD_TIME", format="E", array=mjd_time)
    col2 = fits.Column(name="Rotation", unit="deg", format="E", array=roll_ang)

    send_cols = [col1, col2]
    for star in data_fits.stars:
        send_cols.append(fits.Column(name=star.name, format="E", array=star.photom))
        send_cols.append(
            fits.Column(
                name="FLUX_ERR_{}".format(star.number),
                format="E",
                array=star.uncertainties,
            )
        )
    cols = fits.ColDefs(send_cols)

    hdr = fits.Header()
    hdr["method"] = kwargs["method"]
    hdr["detect"] = kwargs["detect_mode"]
    hdr["initial"] = kwargs["initial_detect"]

    hdr["grid"] = kwargs["grid_bg"]
    hdr["CDPPTYPE"] = kwargs["CDPP_type"]

    primary_hdu = fits.PrimaryHDU(header=hdr)
    hdus.append(primary_hdu)
    hdu = fits.BinTableHDU.from_columns(cols, name="Photometry")
    hdus.append(hdu)

    col1 = fits.Column(
        name="Star", format="B", array=[star.number for star in data_fits.stars]
    )
    col2 = fits.Column(
        name="Cv",
        format="E",
        array=[star.calculate_cdpp(data_fits.mjd_time)[0] for star in data_fits.stars],
    )
    col3 = fits.Column(
        name="Factors", format="E", array=[star.mask_factor for star in data_fits.stars]
    )
    col4 = fits.Column(
        name="Out bounds", format="E", array=[star.out_bound for star in data_fits.stars]
    )
    cols = fits.ColDefs([col1, col2, col3, col4])
    hdu = fits.BinTableHDU.from_columns(cols, name="General")
    hdus.append(hdu)

    hdul = fits.HDUList(hdus)

    hdul.writeto(os.path.join(master_folder, "pyarchi_output.fits"), overwrite=True)

    logger.info("Fit file was created")



import os


def photo_SaveInfo(path, data_fits):
    """
    Stores information related to the photometric run.

    In specific:
        - which file was used
        - mask type
        - detect mode
        - background grid
        - all star related info 
        
    Parameters
    ----------
        path:
            path in which the file will be stored

        data_fits:
             :class:`~pyarchi.data_objects.Data.Data` object.

    Returns
    -------

    """

    with open(os.path.join(path, "photo_info.txt"), mode='w') as file:
        file.write("Used file: {}\n ".format(data_fits.used_file))
        file.write("Mask type: {}\n".format(data_fits.mask_type))
        file.write("Initial mode: {}\n".format(data_fits.init_detection_mode))

        file.write("Detect mode: {}\n".format(data_fits.detect_mode))
        file.write("Background grid: {}\n".format(data_fits.bg_grid))

        file.write("Star \t Cv \t Factors \n")

        for j in data_fits.stars:
            file.write("{} \t {} \t {}\n".format(j.number, j.calculate_cdpp(data_fits.mjd_time)[0], j.mask_factor))



import numpy as np


def create_circular_mask(img, stars, radius, primary, secondary):
    """
    Defines a circular mask for all the stars, with a pre determined radius. If a size_grid_change is different than
    zero
    it converts the mask to that grid size

    Parameters
    ----------
    stars:
          list of all the  :class:`pyarchi.star_track.Star_class.Star` objects
    img:
        First image
    radius:
        radius of the circles. Can be a general radius(same for all the stars) or a dict with different radii,
    in which  the keys are the indexes and the values are the radius

    size_grid_change:
        size of the bigger grid, to which we can convert the mask. If it's zero we don't convert. Otherwise it is
        converted

    primary:
        Methodology to apply to the central star. If it's fits then the initial position of that star is changed to
        be the one determined here.
    secondary:
        Methodology to apply to the outer stars. If it's fits then the initial position of those stars are changed to
        be the ones determined here.


    Returns
    -------
    masks_dict:
        Dictionary where the keys are the number of the star and the values the corresponding mask
    """

    if primary != "circle" and secondary != "circle":
        return {}

    to_calculate = []
    if primary == "circle":
        to_calculate.append(0)

    if secondary == "circle":
        to_calculate = to_calculate + [star.number for star in stars[1:]]

    xx, yy = np.mgrid[: img.shape[0], : img.shape[0]]

    masks_dict = {}
    for index in range(len(stars)):
        if index not in to_calculate:
            continue

        coords = stars[index].positions[0].copy()
        mask = np.zeros(img.shape)

        if not isinstance(radius, (dict)):
            # if the radius if an integer then we use the same value for all stars -> used for optimizing the process
            mask[
                np.where((xx - coords[0]) ** 2 + (yy - coords[1]) ** 2 <= (radius) ** 2)
            ] = 1

        else:
            # If we have an optimized radius we can use it to give value for each star in the list
            mask[
                np.where(
                    (xx - coords[0]) ** 2 + (yy - coords[1]) ** 2
                    <= (radius[str(index)]) ** 2
                )
            ] = 1

        masks_dict[index] = mask

    return masks_dict



import cv2
import numpy as np
from pyarchi.utils import calculate_moments


def create_predictions(Data_fits, img_number):
    """
    Rotates the points in the last know position by the corresponding rotation angle (difference between current angle
     and the one from the last image), predicting the next position of the star

    Parameters
    ----------
    Data_fits:
         :class:`pyarchi.main.initial_loads.Data` object.
    img_number:
         Number of the current image

    Returns
    -------

    """

    predicts = {}
    if img_number + 1 < len(Data_fits.roll_ang):
        fwd_im = Data_fits.roll_ang[img_number + 1]
        bck_im = Data_fits.roll_ang[img_number]

        rot_mat = Data_fits.get_rot_mat(((fwd_im - bck_im) * np.pi / 180), clockwise=True)

        for index in range(len(Data_fits.stars)):
            center = [100, 100]

            if Data_fits.bg_grid != 0:
                scaling_factor = Data_fits.bg_grid / 200
                center = np.multiply(center, scaling_factor) + np.floor(
                    scaling_factor / 2
                )

            pair = Data_fits.stars[index].positions[-1].copy()

            coords = np.dot(rot_mat, [[pair[0] - center[0]], [pair[1] - center[1]]])

            x = coords[0][0] + center[0]
            y = coords[1][0] + center[1]

            predicts[index] = [x, y]

    return predicts


def dynam_method(Data_fits, index, primary, secondary):
    """
    This function is used to calculate the position of the center of each contour. In order to do that
    we calculate the moments of the image, which allows us to derive it's "center of mass".
    All contours with less than 7 points are discarded and, to associate center to star we use the rotate_points
    function to predict the center's expected position. BY comparing the expected positions with the outputs of the
    algorithm we can associate a center to each star.


    Parameters
    ----------
    Data_fits:
         :class:`pyarchi.main.initial_loads.Data` object.
    index
        image's number
    primary:
        Methodology to apply to the central star. If it's dynam then the central star is tracked using this method
    secondary:
        Methodology to apply to the outer stars. If it's dynam then they are tracked using this method

    Returns
    -------

    """

    if primary != "dynam" and secondary != "dynam":
        return

    to_calculate = []
    if primary == "dynam":
        to_calculate.append(0)

    if secondary == "dynam":
        to_calculate = to_calculate + [star.number for star in Data_fits.stars[1:]]

    scaling_factor = Data_fits.bg_grid / 200 if Data_fits.bg_grid != 0 else 1
    if index + 1 < len(Data_fits.roll_ang):

        predictions = create_predictions(Data_fits, index)
        im = Data_fits.get_image(
            index + 1
        )  # prepares the next frame for the detection routine

        centers, _ = calculate_moments(Data_fits.bg_grid, im)

        for detected_center in centers:
            for key, pred_position in predictions.items():
                # calculate x,y coordinate of center
                if np.isclose(
                    detected_center[0], pred_position[0], atol=15 * scaling_factor
                ) and np.isclose(
                    detected_center[1], pred_position[1], atol=15 * scaling_factor
                ):
                    del predictions[key]
                    if Data_fits.stars[key].number not in to_calculate:
                        break

                    Data_fits.stars[key].add_center(detected_center)
                    break
    return 0



import numpy as np


def static_method(Data_fits, img_number, primary, secondary):
    """
    Rotates the points in the last know position by the corresponding rotation angle (difference between current angle
     and the  one from the last image).

    Parameters
    ----------
    Data_fits:
         :class:`pyarchi.main.initial_loads.Data` object.
    img_number:
         Number of the current image
    primary:
        Methodology to apply to the central star. If it's static then the central star is tracked using this method
    secondary:
        Methodology to apply to the outer stars. If it's static then they are tracked using this method

    Returns
    -------

    """
    if primary != "static" and secondary != "static":
        return

    to_calculate = []
    if primary == "static":
        to_calculate.append(0)

    if secondary == "static":
        to_calculate = to_calculate + [star.number for star in Data_fits.stars[1:]]

    if img_number + 1 < len(Data_fits.roll_ang):
        # FIXME:  no need to go get the rotation matrix from the Data class -> get from utils package
        rot_mat = Data_fits.get_rot_mat(
            ((Data_fits.roll_ang[img_number + 1] - Data_fits.roll_ang[0]) * np.pi / 180),
            clockwise=True,
        )

        scaling_factor = Data_fits.bg_grid / 200 if Data_fits.bg_grid != 0 else 1
        delta_pos = [100, 100]
        delta_pos = np.multiply(delta_pos, scaling_factor) + np.floor(scaling_factor / 2)

        for star in Data_fits.stars:

            pair = star.positions[0].copy()

            coords = np.dot(rot_mat, [[pair[0] - delta_pos[0]], [pair[1] - delta_pos[1]]])
            x = coords[0][0] + delta_pos[0]
            y = coords[1][0] + delta_pos[1]

            if star.number in to_calculate:
                star.positions.append([x, y])

    return 0



import numpy as np
from scipy.signal import savgol_filter
from pyarchi.utils import create_logger
import matplotlib.pyplot as plt
logger = create_logger("utils")


def Chunks(l, n, all=False):
    if all:
        jarr = range(0, n - 1)
    else:
        jarr = [0]

    for j in jarr:
        for i in range(j, len(l), n):
            if i + 2 * n <= len(l):
                yield l[i : i + n]
            else:
                if not all:
                    yield l[i:]
                break


def CDPP(flux_vals,times, sized=41, winlen=10, win=30, outl=True):
    """
    Ported version of the K2 CDPP algorithm, implemented by Pedro Silva
    
    Parameters
    ----------
    flux_vals:
        flux values
    times:
        time array
    sized:
        Since we are calculating the CDPP for 1 hour. Window for the Savgol filter
    winlen:
        Convolution window
    win
        Time over which we want to calculate the CDPP

    outl:
        Remove the outliers.

    Returns
    -------

    """
    if flux_vals == []:
        logger.fatal("No values provided")
        return float("nan")

    planeta2 = SavGol2(flux_vals,times, win=sized)

    planeta2 = planeta2 / np.nanmedian(planeta2)

    s = np.r_[
        2 * planeta2[0] - planeta2[winlen - 1 :: -1],
        planeta2,
        2 * planeta2[-1] - planeta2[-1:-winlen:-1],
    ]
    w = np.hanning(winlen)
    planeta3 = np.convolve(w / w.sum(), s, mode="same")
    planeta3 = planeta3[winlen : -winlen + 1]

    planeta4 = planeta2 - planeta3

    M = np.nanmedian(planeta4)
    MAD = 1.4826 * np.nanmedian(np.abs(planeta4 - M))
    out = []
    for i, _ in enumerate(planeta4):
        if (planeta4[i] > M + 5 * MAD) or (planeta4[i] < M - 5 * MAD):
            out.append(i)
    if outl:
        out = np.array(out, dtype=int)
        planeta2 = np.delete(planeta2, out)

    cdpp = 1.0e6 * np.nanmedian(
        [np.std(yi) / np.sqrt(win) for yi in Chunks(planeta2, win, all=False)]
    )  # *1.168
    return cdpp


def Interpolate(time, mask, y):
    """
    Masks certain elements in the array `y` and linearly
    interpolates over them, returning an array `y'` of the
    same length.

    :param array_like time: The time array
    :param array_like mask: The indices to be interpolated over
    :param array_like y: The dependent array

    """

    # Ensure `y` doesn't get modified in place
    yy = np.array(y)
    t_ = np.delete(time, mask)
    y_ = np.delete(y, mask, axis=0)
    if len(yy.shape) == 1:
        yy[mask] = np.interp(time[mask], t_, y_)
    elif len(yy.shape) == 2:
        for n in range(yy.shape[1]):
            yy[mask, n] = np.interp(time[mask], t_, y_[:, n])
    else:
        raise Exception("Array ``y`` must be either 1- or 2-d.")
    return yy


def Scatter(y, win=13, remove_outliers=False):
    """
    Return the scatter in ppm based on the median running standard deviation
    for a window size of :py:obj:`win` = 13 cadences (for K2, this
    is ~6.5 hours, as in VJ14).

    :param ndarray y: The array whose CDPP is to be computed
    :param int win: The window size in cadences. Default `13`
    :param bool remove_outliers: Clip outliers at 5 sigma before computing \
           the CDPP? Default `False`

    """

    if remove_outliers:
        # Remove 5-sigma outliers from data
        # smoothed on a 1 day timescale
        if len(y) >= 50:
            ys = y - Smooth(y, 50)
        else:
            ys = y
        M = np.nanmedian(ys)
        MAD = 1.4826 * np.nanmedian(np.abs(ys - M))
        out = []
        for i, _ in enumerate(y):
            if (ys[i] > M + 5 * MAD) or (ys[i] < M - 5 * MAD):
                out.append(i)
        out = np.array(out, dtype=int)
        y = np.delete(y, out)
    if len(y):
        return 1.0e6 * np.nanmedian(
            [np.std(yi) / np.sqrt(win) for yi in Chunks(y, win, all=True)]
        )
    else:
        return np.nan


def Smooth(x, window_len=100, window="hanning"):
    """
    Smooth data by convolving on a given timescale.

    :param ndarray x: The data array
    :param int window_len: The size of the smoothing window. Default `100`
    :param str window: The window type. Default `hanning`


    """

    if window_len == 0:
        return np.zeros_like(x)
    s = np.r_[2 * x[0] - x[window_len - 1 :: -1], x, 2 * x[-1] - x[-1:-window_len:-1]]
    if window == "flat":
        w = np.ones(window_len, "d")
    else:
        w = eval("np." + window + "(window_len)")
    y = np.convolve(w / w.sum(), s, mode="same")
    return y[window_len : -window_len + 1]


def SavGol(y, win=49):
    """
    Subtracts a second order Savitsky-Golay filter with window size `win`
    and returns the result. This acts as a high pass filter.

    """

    if len(y) >= win:
        return y - savgol_filter(y, win, 2) + np.nanmedian(y)
    else:
        return y
        
def SavGol2(lc,ind,win,withh=False):


	div = (ind[-1]-ind[0])/len(ind)
	maxdiv = div*int(np.sqrt(win/2)+win)
	mindiv = div*int(win-np.sqrt(win/2))
	
	
	Ys = np.zeros((len(lc),win))
	pad = np.pad(lc,(win,win),"symmetric")
	
	metric = np.arange(len(lc))
	for i in range(len(lc)):
		metric[i] = pad[i+win+win//2]-pad[i+win-win//2]
	metric2 = np.nanmedian(np.sqrt(lc)) 
	
	eai = ind[1]-ind[0]
	padi= np.append(np.linspace(ind[0]-win*eai,ind[0],win,endpoint=False),ind)
	padi= np.append(padi,np.linspace(ind[-1],ind[-1]+win*eai,win,endpoint=False))
	

	Id = np.zeros((len(lc),win))
	for i in range(len(lc)):
		Ys[i]+=pad[i+win-win//2:i+win+win//2+1]
		Id[i]+=padi[i+win-win//2:i+win+win//2+1] - padi[i+win]
	
	b = [b[-1]-b[0] for b in Id]
	
	Ks = np.sqrt(Ys)
	
	bgood1 = np.where((b<mindiv)|(b>maxdiv))
	bgood = np.where(metric>metric2)

	if len(bgood1[0])==0 :
		#SavGol can be used
		return SavGol(lc,win)

	elif len(bgood[0])==0:
		#SavGol can be used
		return SavGol(lc,win)
		
	#data set requires SavGol2
	
	
	C = np.zeros((len(lc),win))
	C1 = np.zeros((len(lc),win))

	for i in range(len(lc)):
		A = np.zeros((win,3))

		for j in range(len(A.T)):
			A.T[j] = Id[i]**j				
			
		Kov = np.diag(Ks[i])
		Kov = np.linalg.inv(Kov)
		C[i] = np.linalg.solve(np.dot(np.dot(A.T,Kov),A),np.dot(A.T,Kov))[0]
		C1[i]= np.linalg.solve(np.dot(A.T,A),A.T)[0]


		
	Model = np.sum(C*Ys,axis = 1)
	Model1= np.sum(C1*Ys,axis= 1)
	if withh:
		plt.plot(ind,lc)
		plt.plot(ind,Model)
		plt.show()
	return lc - Model + np.nanmedian(lc)




import glob
import re
import os 

def path_finder(mode, off_curve=None, **kwargs):
    """
    Searches inside the base folder for the desired files

    Parameters
    ----------
    mode:
        - subarray: retrieves the subarray file path
        - default: default lightcurve
        - stars : path for the Star catalogue file
    kwargs:
        - config values

    Returns
    -------

    """

    if mode == "subarray":

        x = glob.glob(os.path.join(kwargs["base_folder"], r"*eduction*/COR/**"))
        cc = " ".join(x)

        return re.compile(r"\S+_Sub\S+.fits").findall(cc)[0]

    elif mode == "default":
        x = glob.glob(os.path.join(kwargs["base_folder"], r"*eduction*/PHE/**"))
        cc = " ".join(x)

        curve_2_search = kwargs["official_curve"] if not off_curve else off_curve
        return re.compile(r"\S+{}\S+.fits".format(curve_2_search)).findall(cc)[0]

    elif mode == "stars":
        x = glob.glob(os.path.join(kwargs["base_folder"], r"CH_*/data/**"))

        cc = " ".join(x)
        return re.compile(r"\S+Star\S+.fits").findall(cc)[0]



import numpy as np
from cv2 import findContours, moments, threshold


def calculate_moments(bg_grid, img):
    """
    Apply image moments, as described in `OpenCv docs <https://docs.opencv.org/3.4.2/dd/d49/tutorial_py_contour_features.html>`_

    This functions pre-processes the image before calculating the moments, so that it conforms to OPenCv's input data types.   
    Parameters
    ----------
    bg_grid : int
        Size of the background grid
    img : np.array
        Image to be processed
    
    Returns
    -------
    centers
        Position of each detected star
    distances
        Distance of each star to the image center -> position [100,100]
    """

    scaling_factor = bg_grid / 200 if bg_grid else 1
    im = img.copy()
    im /= np.nanmax(im)
    im *= 255
    im = np.uint8(im)
    _, thresh = threshold(im, 10, 255, 0)
    contours, _ = findContours(thresh, 1, 2)

    new_cont = []
    for j in reversed(contours):  # removes small contours
        if len(j) > 10:
            new_cont.append(j)
    contours = new_cont
    del new_cont
    centers = []
    distances = []

    center = [100, 100]
    if bg_grid != 0:
        center = np.multiply(center, scaling_factor) + np.floor(scaling_factor / 2)

    for j in range(len(contours)):
        M = moments(contours[j])

        cY = M["m10"] / M["m00"]
        cX = M["m01"] / M["m00"]

        centers.append([cX, cY])
        distances.append(np.sqrt((cX - center[0]) ** 2 + (cY - center[1]) ** 2))

    return centers, distances



from astropy.io import fits
import numpy as np

from .Mask import Masks

from pyarchi.utils import path_finder
from pyarchi.utils import create_logger, CDPP, DRP_CDPP

logger = create_logger("star_track")


class Star:
    """
    Star class that holds the star information. We have a class attribute, number, that will be used to name the star.
     Each star increments this number by one.

    Parameters
    ---------------
        cdpp_type:
            CDPP algorithm 
            
        pos=None:
            Initial position
            
        dist = None:
            Distance to center
    Notes
    -----------
        name:
            Designation of the star
        masks:
            Holds the mask used in each image.
        GP_data:
            Holds the GPs results
        positions:
            position, in pixels, of the centroid. Always on coordinates of the [200,200] grid
        init_pos:
            Initial position, in pixels, of the centroid. Always on coordinates of the [200,200] grid

        out_bound:
            An overlap between the mask and the empty region was found

        _active:
            Do we want to calculate the flux for this star

        photom:
            holds the light curve information
        debug:
            If 1 we compare the data obtained from our analysis with the one produced by the official pipeline
    """

    number = 0

    def __init__(self, cdpp_type, pos=None, dist = None):

        # General star information
        self.number = Star.number
        self.__class__.number += 1
        self.positions = [pos]
        self.init_pos = pos.copy()
        self.dist_center = dist  
        
        # Star status:
        self.out_bound = False  # True if the mask overlaps the empty zone
        self._active = True

        # Data storage
        self.masks = None
        self._GP_data = None

        # misc
        self._photom = []
        self._uncertainties = []
        self.debug = 0
        self.default_lightcurve = []

        self.DRP_uncert = None

        self.cdpp_type = cdpp_type

    def enable_debug(self, **kwargs):
        """
        Extracts the default data from  the FITS files
        Parameters
        ----------
        base_folder:
            path to folder in which the FITS files are located

        Returns
        -------

        """

        try:
            logger.info("Sepyarchng for the file")
            default_path = path_finder(mode="default", **kwargs)
        except:
            logger.fatal("Could not find the path to the file", exc_info=True)
            return -1

        with fits.open(default_path) as hdulist:
            self.default_lightcurve = hdulist[1].data["flux"]
            self.DRP_uncert = hdulist[1].data["FLUXERR"]
        self.debug = 1

    def calculate_cdpp(self, time=None):
        """
        Calculates the CDPP of the light curve, in order to quantify the results

        Returns
        -------
            CDPP, CDPP_def : noise for my light curve and for the official one, respectively
        """

        # TODO: allow to pass values for the CDPP
        if not self.is_active:  # If not active return stupid values
            return 2e7, 2e7

        sized = 41
        winlen = 10
        win = 30
        outl = True

        if time is None:
            logger.fatal("Missing the time to use DRP's CDPP")
        
        if self.cdpp_type == "K2":
            cv = CDPP(self.photom,time, sized, winlen, win, outl)
            cv_def = CDPP(self.default_lightcurve, time, sized, winlen, win, outl) if self.debug else None

        elif self.cdpp_type == "DRP":

            cv = DRP_CDPP(self.photom.copy(), time, 0.5 / 24)
            cv_def = DRP_CDPP(self.default_lightcurve, time, 0.5 / 24) if self.debug else None


        return cv, cv_def

    def add_initial_mask(self, mask, factor, scaling_factor, low_memory=0):
        """
        Initializes the Masks class

        Parameters
        ----------
        mask
            Initial mask for the star
        factor
            Increase factor for the mask
        size_grid_change
            Size of the background grid
        low_memory:
            Mode in which only the necessary information is kept during the process
        Returns
        -------
            0
                If no error is found
            -1
                If the factor is a negative number
        """
        logger.info("{} - Adding initial masks".format(self.name))
        if type(factor) == dict:
            self.masks = Masks(factor[str(self.number)], scaling_factor, mask, low_memory)
            return 0
        if factor <= 0:
            logger.fatal("Invalid mask factor")
            return -1
        self.masks = Masks(factor, scaling_factor, mask, low_memory)

        return 0

    def change_init_pos(self, pos):
        """
        Change the star's initial position. Used when more than one initial center determination method is active
        and some information needs to be overwritten
        Parameters
        ----------
        pos:
            New position for the centroid of the star (for the 1st image)

        Returns
        -------

        """
        self.positions = [pos]
        self.init_pos = pos.copy()

    def update_mask(self, scaling_factor, index):
        """
        Updates the mask with the information from the current image
        Parameters
        ----------
        scaling_factor
            Size of the background grid.
        index
            Index of the current image

        Returns
        -------

        """

        if index == 0:
            return

        init_pos = self.init_pos
        positions = self.positions[-1]

        x_change = int(round(-init_pos[0] + positions[0]))
        y_change = int(round(-init_pos[1] + positions[1]))

        self.masks.update_mask(x_change, y_change, index)

    def remove_data(self):
        """
        Empty all information stored on this star. Used during the optimization process
        Returns
        -------

        """
        logger.info("Removing all data")
        del self.masks

        self.masks = None
        self.positions = [self.init_pos.copy()]
        self._photom = []
        self._uncertainties = []
        self.out_bound = False  # True if the mask uses the empty zone
        self._GP_data = None

    def calculate_errors(self, background, nstack, cron, dark):
        flux = self.photom[-1]
        npix = self.masks.normalized_points
        self._uncertainties.append(
            np.sqrt(flux + background * npix + npix * nstack * cron ** 2 + dark * npix)
        )

    def add_mask(self, mask):
        """
        Adds a mask to the Masks object

        Parameters
        ----------
        mask:
            mask to add

        Returns
        -------

        """

        self.masks.add_mask(mask)

    def add_center(self, point):
        self.positions.append(point)

    def change_center(self, index, new_point):
        self.positions[index] = new_point

    def disable(self):
        """
        Sets the flag to disable the calculation of the flux of this star.
        Returns
        -------

        """
        logger.info("Star {} was disabled".format(self.number))
        self._active = False

    def import_photom(self, photom):
        """
        Used when loading data from disk
        Parameters
        ----------
        photom

        Returns
        -------

        """
        self._photom = photom

    def all_masks(self):
        return self.masks.all

    def import_uncertainties(self, uncerts):
        self._uncertainties = uncerts

    @classmethod
    def reset_number(cls,new_number):
        cls.number = new_number

    @property
    def name(self):
        return "Star {}".format(self.number)
    # Properties:
    @property
    def is_active(self):
        return self._active

    @property
    def photom(self):
        return self._photom

    @property
    def uncertainties(self):
        return (
            self._uncertainties if self._uncertainties else [500e-6 for _ in self.photom]
        )

    @property
    def mask_factor(self):
        return self.masks.factor

    @property
    def mask_npoints(self):
        return self.masks.size

    @property
    def mask_norm_npoints(self):
        return self.masks.normalized_points


    @property
    def first_mask(self):
        return self.masks.first

    @property
    def latest_mask(self):
        """
        Returns the last mask added to the star.
        Returns
        -------

        """
        return self.masks.latest

    def __str__(self):
        return "{} - {}".format(self.name, self._active)

    def __repr__(self):
        return "{} - {}".format(self.name, self._active)



import logging
import sys
from logging.handlers import TimedRotatingFileHandler

FORMATTER = logging.Formatter("%(name)s — %(levelname)s — %(message)s")
LOG_FILE = "my_app.log"


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    file_handler = TimedRotatingFileHandler(LOG_FILE, when="midnight")
    file_handler.setFormatter(FORMATTER)
    return file_handler


def create_logger(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)  # better to have too much log than not
    # enough
    if logger.handlers:
        logger.handlers = []

    logger.addHandler(get_console_handler())
    logger.addHandler(get_file_handler())
    # with this pattern, it's rarely necessary to propagate the error up to
    # parent
    logger.propagate = False
    return logger



import numpy as np
from scipy import optimize


def matrix_cnter_clock(angle):
    """
    Creates a rotation matrix for counter clockwise rotations.
    Expects an angle in degrees

    :param angle: rotation angle, in degrees.
    :return: counter clockwise rotation matrix
    """
    return np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])


def matrix_clockwise(angle):
    """
    Creates a rotation matrix for clockwise rotations.
    Expects an angle in degrees

    :param angle: rotation angle, in degrees.
    :return: counter clockwise rotation matrix
    """
    return np.array([[np.cos(angle), np.sin(angle)], [-np.sin(angle), np.cos(angle)]])


if __name__ == "__main__":
    pass



# Decorators
from functools import wraps


def my_timer(orig_func):
    """
    Decorator to measure the time spent on the function
    Parameters
    ----------
        orig_func:
            FUnction to be analysed

    Returns
    -------

    """
    import time

    @wraps(orig_func)
    def wrapper(*args, **kwargs):
        t1 = time.time()
        result = orig_func(*args, **kwargs)
        t2 = time.time() - t1
        if not kwargs["optimize"]:
            print("{} ran in: {} sec".format(orig_func.__name__, t2))
        return result

    return wrapper



from matplotlib import pyplot as plt
import os
import numpy as np


from pyarchi.utils import create_logger

logger = create_logger("Output creation")


def photom_plots(data_fits, master_folder, singular=None, **kwargs):
    """
    Processes the data collected from the images and outputs the desired graphs.

    Parameters
    ----------
    data_fits:
        :class:`~pyarchi.data_objects.Data.Data` object with all the stars information inside
    singular:
        If it's not None, then only that star's light curve is saved. Only works for the "show_results".
    master_folder:
        Root folder in which all the data shall be stored
        
    kwargs:
        Configuration values that are used to change the data processed and the output data:

    Returns
    -------

    """

    if kwargs["headless"]:
        plt.switch_backend("Agg")

    scaling_factor = kwargs["grid_bg"] / 200 if kwargs["grid_bg"] else 1

    logger.info("Starting the plot process")
    point_size = 3  # size of the dots on the graphs

    colors1 = plt.get_cmap("jet_r")
    x = data_fits.mjd_time

    if kwargs["report_pictures"]:
        fig, ax = plt.subplots(2, 2)
        fig_1, ax_1 = plt.subplots()

        pos = [[0, 0], [0, 1], [1, 0], [1, 1]]

        colors1 = plt.get_cmap("jet_r")

        for j in range(len(data_fits.stars)):
            cdpp = data_fits.stars[j].calculate_cdpp(data_fits.mjd_time)[0]

            ax[pos[j][0], pos[j][1]].scatter(
                x,
                data_fits.stars[j].photom,
                color="black",
                label="Cv = {:.2f}".format(cdpp),
                s=point_size,
            )

            ax[pos[j][0], pos[j][1]].set_title(data_fits.stars[j].name)
            ax[pos[j][0], pos[j][1]].legend(loc=4)

            ax[pos[j][0], pos[j][1]].ticklabel_format(axis="y", style="sci")
            ax[pos[j][0], pos[j][1]].yaxis.major.formatter.set_powerlimits((0, 0))

            # Figure 2:
            ax_1.scatter(
                x,
                data_fits.stars[j].photom,
                c=colors1((5 * j) / 40),
                label=data_fits.stars[j].name,
                zorder=-j,
                s=point_size,
            )

            ax_1.ticklabel_format(axis="y", style="sci")
            ax_1.yaxis.major.formatter.set_powerlimits((0, 0))

        ax_1.legend()
        fig.tight_layout()
        fig_1.tight_layout()

    fig_2, ax_2 = plt.subplots()

    img = data_fits.get_image(0)
    ax_2.imshow(img)
    circles = []
    names = []
    [
        logger.info(
            star.name + " : {} ppm".format(star.calculate_cdpp(data_fits.mjd_time)[0])
        )
        for star in data_fits.stars
    ]
    for j, star in enumerate(data_fits.stars):
        circle_rad = 10
        if kwargs["grid_bg"]:
            circle_rad = int(circle_rad * scaling_factor)

        positions = star.positions[0]
        circle1 = plt.Circle(
            (positions[1], positions[0]),
            circle_rad,
            edgecolor=colors1((5 * j) / 40),
            facecolor="none",
        )

        ax_2.add_artist(circle1)
        circles.append(circle1)
        names.append(data_fits.stars[j].name)
    ax_2.legend(circles, names)

    fig_2.savefig(os.path.join(master_folder, "star_names"))

    if kwargs["debug"]:
        data_fits.stars[0].enable_debug(**kwargs)
        cv, cv_def = data_fits.stars[0].calculate_cdpp(data_fits.mjd_time)
        fig = plt.figure()

        plt.scatter(
            x,
            data_fits.stars[0].default_lightcurve
            / np.max(data_fits.stars[0].default_lightcurve),
            label=kwargs["official_curve"],
            s=point_size,
        )

        plt.scatter(
            x,
            data_fits.stars[0].photom / np.max(data_fits.stars[0].photom),
            label=kwargs["method"],
            s=point_size,
        )

        plt.title(" Normalized lightcurves")
        plt.legend()
        fig_1 = plt.figure()

        plt.scatter(
            x,
            data_fits.stars[0].default_lightcurve,
            label="DRP; Cv = {}".format(round(cv_def, 2)),
            s=point_size,
        )
        plt.scatter(
            x,
            data_fits.stars[0].photom,
            label="pyarchi; Cv = {}".format(round(cv, 2)),
            s=point_size,
        )
        plt.title("Lightcurves")
        plt.legend(bbox_to_anchor=(0.6, 1), ncol=1)

        fig.savefig(os.path.join(master_folder, "Star_0/normalized_curves"))
        fig_1.savefig(os.path.join(master_folder, "Star_0/pyarchi_official_compare"))

    if kwargs["show_results"]:
        # PLot the ligthcurve
        for j in range(len(data_fits.stars)):

            if singular is not None and singular != j:
                continue

            plt.figure()
            cv, _ = data_fits.stars[j].calculate_cdpp(data_fits.mjd_time)

            plt.scatter(
                x,
                data_fits.stars[j].photom,
                label="Method - {}; Cv = {}".format(kwargs["method"], round(cv, 2)),
                s=point_size,
            )
            plt.title("Lightcurve of {}".format(data_fits.stars[j].name))
            plt.legend(loc=1)
            plt.savefig(os.path.join(master_folder, "Star_{}/pyarchi_curve".format(j)))

    logger.info("All plots were completed")



import os


def parameters_validator(**kwargs):
    """
    Loads the configuration parameters from the .yaml file. After loading,
    it  checks some of the parameters to see if they have valid values.

    Parameters
    ----------
    parameters

    Returns
    -------

    """
    wrong_params = []
    warnings = []
    paths_to_test = ["base_folder", "optimized_factors", "results_folder"]

    for key in paths_to_test:
        if not os.path.exists(kwargs[key]):
            wrong_params.append(key)

    if kwargs["official_curve"] not in ["DEFAULT", "OPTIMAL", "RINF", "RSUP"]:
        wrong_params.append(kwargs["official_curve"])

    for param in ["method", "detect_mode", "initial_detect"]:
        if type(kwargs[param]) is not str:
            wrong_params.append(param)
            continue

        if "+" in kwargs[param]:
            modes = kwargs[param].split("+")
        else:
            modes = [kwargs[param]]

        if param == "method":
            valid_values = ["circle", "shape"]
        if param == "detect_mode":
            valid_values = ["dynam", "offsets", "static"]
        elif param == "initial_detect":
            valid_values = ["dynam", "fits"]

        for md in modes:
            if md not in valid_values and param not in wrong_params:
                wrong_params.append(param)

    if not isinstance(kwargs["val_range"], list):
        if kwargs["optimize"]:
            wrong_params.append("val_range")
        else:
            warnings.append("val_range")
    else:
        optim_vals = kwargs["val_range"]
        if optim_vals[1] <= optim_vals[0] < 0 or optim_vals[1] < 0:
            if kwargs["optimize"]:
                wrong_params.append("val_range")
            else:
                warnings.append("val_range")

    if (kwargs["grid_bg"] / 200) % 2 != 1 and kwargs["grid_bg"] != 0:
        wrong_params.append("grid_bg")

    return wrong_params, warnings, kwargs



import numpy as np
import cv2

from pyarchi.data_objects import Star
from pyarchi.utils import calculate_moments


def initial_dynam_centers(img, bg_grid, **kwargs):
    """

    Uses the same method as the one in dynamical_centers to detect the original position of each star.
    Afterwards, the positions are ordered by closeness to the center which allows us to keep the same
    order as in the "fits" initial position method.
    Parameters
    ----------
    im:
        First image in the data set
    bg_grid:
        Size of the background grid
    Returns
    -------

    """

    centers, distances = calculate_moments(bg_grid, img)
    stars = []

    for dist in sorted(distances):
        positions = centers[distances.index(dist)]
        stars.append(Star(kwargs["CDPP_type"], positions, dist))
    
    return stars


if __name__ == "__main__":
    pass




import cv2
import numpy as np

from pyarchi.data_objects.Mask import logger


def create_shape_mask(im, stars, increase_factor, scaling_factor, primary, secondary):
    """
    Finds the contours of the image, with openCv default functions

    Parameters
    ----------
    im:
        copy of the image used for the shape detection
    stars:
        list of all the  :class:`pyarchi.star_track.Star_class.Star` objects

    increase_factor:
        Number of pixels added to the outside of the shape. For example, if factor = 1 then we add a layer of pixels
        around the entire shape

    size_grid_change:
        SIze of the background grid in use

    primary:
        Methodology to apply to the central star. If it's dynam then the initial position of that star is changed to
        be the one determined here.
    secondary:
        Methodology to apply to the outer stars. If it's dynam then the initial position of those stars are changed to
        be the ones determined here.

    Returns
    -------
    masks_dict:
        Dictionary where the keys are the number of the star and the values the corresponding mask
    """

    if primary != "shape" and secondary != "shape":
        return {}

    to_calculate = []
    if primary == "shape":
        to_calculate.append(0)

    if secondary == "shape":
        to_calculate = to_calculate + [star.number for star in stars[1:]]

    # Normalization to use with OpenCv
    im /= np.nanmax(im)
    im *= 255
    im = np.uint8(im)

    # TODO: change this threshold
    _, thresh = cv2.threshold(im, 10, 255, 0)
    contours, _ = cv2.findContours(thresh, 1, 2)

    to_remove = []
    for j in range(len(contours)):  # removes small contours (under 5 points)
        if len(contours[j]) <= 11:
            to_remove.append(j)
    for rm in reversed(to_remove):
        contours.pop(rm)
    
    if len(contours) != len(stars):  # we need to have the same number of masks and stars
        logger.fatal("Number of detected contours and stars does not add up")
        logger.fatal(" \t Contours: {}; Stars:{}".format(len(contours),len(stars)))

        return -1

    mask_dict = {}
    for cont in contours:
        mask = np.zeros(im.shape)
        cv2.drawContours(
            mask, [cont], -1, (255, 255, 255), -1
        )  # fills the contour with data

        mask[np.where(mask != 0)] = 1  # normalizes the array

        for index, star in enumerate(stars):
            if index not in to_calculate:
                continue
            pos = star.init_pos.copy()

            if mask[int(round(pos[0])), int(round(pos[1]))] != 0:
                # since the contours are not ordered like the stars one must check if we have data from the contours in
                # the specified position

                if isinstance(increase_factor, (np.int64, int)):
                    final_mask = shape_increase(mask, increase_factor)
                else:
                    final_mask = shape_increase(mask, increase_factor[str(star.number)])
                mask_dict[index] = final_mask

    return mask_dict


def shape_increase(data, factor, fac=1):
    """
    Increases the boundary of the mask by one pixel. i.e., adds one layer of pixels around the mask present in the data
    passed in.

    Parameters
    --------------
    data:
        Array with the original shape that we wish to expand
    factor:
        number of pixels that we wish to increase
    fac:
        Used internally

    Returns
    -------
        numpy array:
            Increased image
    Notes
    -----

        This is a recursive function to expand by a number of pixels (factor) our image inside the data array.
        This function muss be refactored since it's inefficient. However, for the time being, the overhead that it introduces
        is not enough to justify optimizing it.
    """
    # ToDO: refactor this. IN other words: prettify this
    if factor < 1:
        return data

    positions = np.where(data != 0)  # array positions that we wish to increase

    new = np.zeros(data.shape)

    cases = [[0, 0], [0, 1], [1, 0], [1, 1], [0, -1], [-1, 0], [-1, -1], [1, -1], [-1, 1]]

    max_pos = data.shape[0]
    for pos in zip(positions[0], positions[1]):

        for j in cases:

            val_x = j[0]
            val_y = j[1]

            if pos[0] + val_x < 0:
                val_x = -pos[0]
            elif pos[0] + val_x >= max_pos:
                val_x = max_pos - pos[0] - 1

            if pos[1] + val_y < 0:
                val_y = -pos[1]

            elif pos[1] + val_y >= max_pos:
                val_y = max_pos - pos[1] - 1

            new[pos[0] + val_x, pos[1] + val_y] = 1

    if fac >= factor:
        return new
    elif fac > 1000:
        # Cap of 1000 iterations
        logger.fatal("Iteration cap was reached. Problems increasing the mask")
        return -1
    else:
        return shape_increase(new, factor, fac + 1)



import numpy as np

from astropy.io import fits
from pyarchi.utils import path_finder
from pyarchi.utils.misc.rotation_mats import matrix_cnter_clock
from pyarchi.utils import create_logger
from pyarchi.data_objects import Star

logger = create_logger("initial_detection")

def centers_from_fits(primary, secondary, stars, initial_angle,initial_offset, **kwargs):
    """
    Using information stored on the fits files, we determine the centers positions. The centers are determined using
    relations between the differences in RA and DEC of all stars in relation to the known point : the central star.

    After determining the center, we use the primary and secondary arguments to see if this function should change
    the initial position of the star.

    Parameters
    ----------
    primary: str
        Methodology to apply to the central star. If it's fits then the initial position of that star is changed to
        be the one determined here.
    secondary: str
        Methodology to apply to the outer stars. If it's fits then the initial position of those stars are changed
            to be the ones determined here.
    stars: list
        List with all the stars found with the dynam method
    initial_angle: float
        Rotation angle of the satellite for the first image
    initial_offset: list    
        DRP's estimation of the central star location
    kwargs
        kwargs
        
    Returns
    -------
        List:
            Updated list of stars, with the positions determined by the fits method

    """

    if primary != "fits" and secondary != "fits":
        return stars

    to_calculate = []

    if primary == "fits":
        to_calculate = to_calculate + [0]

    if secondary == "fits":
        to_calculate = to_calculate + [1]

    arcsec = 1 / 3600
    arcsec_per_pix = 1
    scaling_factor = kwargs['grid_bg'] / 200 if kwargs['grid_bg'] != 0 else 1

    r_ang = initial_angle

    stars_path = path_finder(mode="stars", **kwargs)

    try:
        hdulist = fits.open(stars_path)
    except IOError:
        logger.error("Star Catalogue file not found")
        return -1
    else:
        with hdulist:
            cent_ra = hdulist[1].header["CENT_RA"]
            cent_dec = hdulist[1].header["CENT_DEC"]
            first_RA = hdulist[1].data["RA"]
            first_DEC = hdulist[1].data["DEC"]
            mags = hdulist[1].data["MAG_CHEOPS"]

    logger.info(
        "Extracted StarCatalogue's information; Starting the position analysis process"
    )

    central_star = initial_offset

    valid_stars = (
        0
    )  # number of valid stars found so far. There are more than the usable ones
    distances = []
    positions = []

    if 0 in to_calculate:  # for the central star
        pos = [central_star[1] - 412, -412 + central_star[0]]
        pos = np.multiply(pos, scaling_factor) + np.floor(scaling_factor / 2)
        stars[valid_stars].change_init_pos(pos)

    for index in range(len(first_DEC)):
        star_ra, star_dec = first_RA[index], first_DEC[index]
        r_mat = matrix_cnter_clock((360 - r_ang) * np.pi / 180)

        # http://spiff.rit.edu/classes/phys373/lectures/astrom/astrom.html
        delta_RA = (cent_ra - star_ra) * np.cos((cent_dec * np.pi / 180)) / arcsec
        delta_DEC = (cent_dec - star_dec) / arcsec

        initial_coords = [[delta_RA], [delta_DEC]]
        new_coords = np.dot(r_mat, initial_coords)

        star_ra = new_coords[0][0]
        star_dec = new_coords[1][0]

        x_pos = 100 - (central_star[1] - 512 + arcsec_per_pix * star_dec)
        y_pos = 100 + central_star[0] - 512 + arcsec_per_pix * star_ra

        if 0 < x_pos < 200 and 0 < y_pos < 200:
            if mags[index] <= 13:
                distances.append(
                    np.sqrt(
                        (x_pos - (central_star[0] - 412)) ** 2
                        + (y_pos - (central_star[1] - 412)) ** 2
                    )
                )
                positions.append([x_pos, y_pos])

    valid_stars = 1
    warning_mismatch = 0
    if len(distances) != len(stars):
        logger.warning("Mismatch between fits and dynam initial detection methods. Possible non-visible stars inside the image!!")
        logger.warning("Creating all of the background stars from fits file; Disregarding data from 'dynam' method")
        warning_mismatch = 1
        stars = stars[:1]
        Star.reset_number(1)

    for dist in sorted(distances[1:]):
        pos = positions[distances.index(dist)]

        if 1 in to_calculate:  # for all other stars
            pos = np.multiply(pos, scaling_factor) + np.floor(scaling_factor / 2)

            if warning_mismatch: # if the fits and dynam do not match, create from scratch
                stars.append(Star(kwargs["CDPP_type"], positions, dist))
            else:
                stars[valid_stars].change_init_pos(pos)
                valid_stars += 1

    return stars


import numpy as np
from pyarchi.utils import create_logger

logger = create_logger("masks")


class Masks:
    """
        Class used to hold the mask for each iteration, as well as some of some important methods
    """

    def __init__(self, mask_factor, grid_increase, initial_mask, low_memory=0):

        self._mask_factor = mask_factor
        self.grid_increase = grid_increase

        self._masks = []
        self._low_mem = low_memory

        self._mask_size = len(np.where(initial_mask != 0)[0])
        self.add_mask(initial_mask)

    def add_mask(self, mask):
        """
        Stores a new mask. If the low memory mode is active, only two masks are stored in memory: the initial one
        and the latest.

        Parameters
        ----------
        mask:
            Mask to be added

        Returns
        -------

        """

        if self._low_mem and len(self._masks) == 2:
            self._masks[-1] = mask
        else:
            self._masks.append(mask)

    def update_mask(self, x_change, y_change, image_number):
        """
            Changes the mask to the correct position in the new image. If the grid_bg is not None, then
            the calculations are made with the bigger grid.

        Parameters
        ----------
        x_change
            change in the x direction
        y_change
            change in the y direction

        image_number
            Current image number.

        Returns
        -------

        """

        empty_mask = np.roll(self.first.copy(), x_change, axis=0)
        empty_mask = np.roll(empty_mask, y_change, axis=1)

        if image_number != 0:  # no need to add again the mask for the first frame
            self.add_mask(empty_mask)

    @property
    def latest(self):
        """
            Returns the latest mask stored in the class
        """
        return self._masks[-1]

    @property
    def first(self):
        """
            Returns the first mask
        """
        return self._masks[0]

    @property
    def number_masks(self):
        """
        Return the number of masks stored in the obejct
        """
        return len(self._masks)

    @property
    def all(self):
        """
        Return a list with all of the stored masks
        """
        return self._masks.copy()

    @property
    def factor(self):
        """
            Returns the mask's size  (i.e. circle radius or layers of pixels added to the mask)
        """
        return self._mask_factor

    @property
    def size(self):
        """ 
            Returns the number of pixels in the mask
        """
        return self._mask_size

    @property
    def normalized_points(self): 
        """ 
            Returns the number of points, normalized to the "normal" grid, with 200 by 200 px
        """
        return self._mask_size / (self.grid_increase) ** 2

    def __getitem__(self, item):
        return self._masks[item]  # delegate to li.__getitem__



import numpy as np


def DRP_CDPP(photom, jd, twidth):
    """
    Compute the Combined Differential Photometric Precision (CDPP) metric.
    This definition estimates the unbiased variance of a rolling window of size twidth
    (in sample units) and returns the sqrt( meadian (var) ) / sqrt (norm) where norm
    is the average number of points in the windows.
    This implementation is robust against gaps in the light curve.

    @author Sergio Hoyer, Pascal Guterman

    Parameters
    ------------------
        flux : Flux. 1d ndarray time series.
        jd : Timescale (in days). 1d ndarray of same size.
        twidth: temporal width where the CDPP is estimated (same units as JD)


    Returns
    --------------
        CDPP: SQRT ( MEDIAN ( VAR_ARRAY ) ) / SQRT(NORM)
        WINDOW_SIZE: CONSTANT. Number of points in rolling windows


    Syntax
    ---------
        cdpp, wsize = cdpp(flux, jd, twidth)

    Example
    -------------
    """
    if photom == []:
        print("No values provided for CDPP")
        return float("nan")

    flux = 1e6 * (photom / np.mean(photom) - 1)  # normalized flux on ppm

    # Convert to sample numbers
    texp = np.median(np.diff(jd))  # Exposure time
    nexp = len(flux)  # Number of points in the LC
    # Evenly spaced signal
    index = np.round((jd - np.min(jd)) / texp).astype(int)
    eflux = np.ones(1 + np.max(index)) * np.nan

    try:
        eflux[index] = flux
    except Exception as e:
        print(jd)
        print(type(photom))
        print(photom)
        print(index)
        print(eflux)
        raise (e)

    xwidth = np.round(twidth / texp).astype(int)  # Requested time width (in sample unit)
    wsize = np.clip(
        xwidth, 2, nexp
    )  # Sliding window, minimum length 2 points / max full curve

    # To mute the warning rise by numpy nanvar when the rolling window is full of nans (gaps)
    with np.warnings.catch_warnings():
        np.warnings.filterwarnings("ignore", category=RuntimeWarning)

        # array with the variance on each rolling window using unbiased estimator (N-1)
        vararr = [
            np.nanvar(eflux[i : i + wsize], ddof=1) for i in range(nexp - wsize + 1)
        ]

    # Convention length param = 0 returns point to point noise

    # Normalization factor of the standard deviation
    norm = np.clip(xwidth, 1, nexp)

    return np.sqrt(np.nanmean(vararr)) / np.sqrt(norm)


if __name__ == "__main__":
    pass



import matplotlib.pyplot as plt
import os

from pyarchi.utils import export_txt, create_fits, photo_SaveInfo
from pyarchi.data_objects import Data

from .photometry_outputs import photom_plots
from .folder_handler import handle_folders

from pyarchi.utils import create_logger
logger = create_logger("Output creation")

def store_data(data_fits, job_number, singular=None, **kwargs):
    """
    Stores the data extracted from the entire pipeline.

    Like the controllers, it can work with a :class:`~pyarchi.data_objects.Data.Data` object that was used for both parts
    or only for one of them.

    If one wishes, it can also only create the folders and extract the data only for the chosen star.
   
    Parameters
    ----------
    data_fits:
        :class:`~pyarchi.data_objects.Data.Data` object.
    job_number:
        Job number, from the supernova server. Will be used to save the parent folder inside the
        kwargs["results_folder"] to create the entire directory structure.
    singular:
        If it's not None, then only that star's light curve is saved. Only works for the "show_results".

    kwargs: 
        Dictionary with the configuration values. Can be obtained with the  :class:`~pyarchi.routines.Photo_Controller.Photo_controller` parameters property
    """
    if kwargs["headless"]:
        plt.switch_backend("Agg")

    if not isinstance(data_fits, Data):
        logger.fatal("Data object found an error during run time. No data will be stored")
        return -1

    if data_fits.abort_process:
        logger.fatal("Data object came from an aborted run. No data will be extracted")
        raise RuntimeError(
            "Data object came from an aborted run. No data will be extracted"
        )
    if singular is not None and not isinstance(singular, int):
        logger.fatal("the 'singular' parameter should be an integer")
        raise ValueError("the 'singular' parameter should be an integer")

    master_folder = handle_folders(len(data_fits.stars), job_number, **kwargs)

    if not data_fits.is_empty:
        if data_fits.low_mem:
            data_fits.reload_images(**kwargs)
        photo_SaveInfo(master_folder, data_fits)
        photom_plots(data_fits, master_folder, singular, **kwargs)

        if kwargs["export_text"]:
            logger.info("Starting txt file creation")
            if export_txt(data_fits, master_folder, **kwargs) == -1:
                logger.fatal("File creation has failed")

        if kwargs["export_fit"]:
            logger.info("Starting fits file creation")
            if create_fits(master_folder, data_fits, **kwargs) == -1:
                logger.fatal("Problems were found during fit file creation")

    if kwargs["show_results"]:

        plt.show()

    plt.close("all")



from astropy.io import fits
import numpy as np
import os

from pyarchi.utils import get_optimized_mask, create_logger
from pyarchi.utils import path_finder

logger = create_logger("util")


def export_txt(Data_fits, path, **kwargs):
    """
    Stores the light curves on a text file with the following format:
        MJD_TIME;ROLL_ANGLE FLux <Star i> FLUX_ERR <Star i>

    Parameters
    ----------
    Data_fits
        :class:`~pyarchi.data_objects.Data.Data` object. 
    path:
        Path in which the file shall be stored
    kwargs
        Configuration values

    Returns
    -------

    """

    default_path = path_finder(mode="default", **kwargs)
    file_name = "/" + kwargs["base_folder"].split("/")[-2].split("/")[-1]
    try:
        hdulist = fits.open(default_path)
    except IOError:
        logger.error("Lightcurve-Default file not found", exc_info=True)
        return -1
    else:
        with hdulist:
            roll_ang = hdulist[1].data["ROLL_ANGLE"]
            mjd_time = hdulist[1].data["MJD_TIME"]

    off_curve = default_path.split("-")[-1].split("_")[0]

    with fits.open(default_path) as hdulist:
        default_lightcurve = hdulist[1].data["flux"]
        default_err = hdulist[1].data["FLUXERR"]

    header = "mjd_time; roll_ang; {} lightcurve".format(off_curve)
    np.savetxt(
        path + file_name + ".txt",
        np.c_[mjd_time, roll_ang, default_lightcurve, default_err],
        delimiter=" ",
        header=header,
    )

    results_file = os.path.join(path, "pyarchi_output.txt")

    header = "Method - {}\n".format(kwargs["method"])
    header += "Detect mode - {}\n".format(kwargs["detect_mode"])
    header += "Initial load - {}\n".format(kwargs["initial_detect"])
    header += "Background grid - {}\n".format(kwargs["grid_bg"])
    optim_mask = get_optimized_mask(**kwargs)
    header += "Stars information:\n"
    for key, val in optim_mask.items():
        header += "\t Star: {}; Factor: {}; Out of bounds: {}\n".format(
            key, val, Data_fits.stars[int(key)].out_bound
        )

    header += "mjd_time; roll_ang; -> Stars"

    a = Data_fits.all_curves
    b = Data_fits.all_uncertainties
    c = np.empty((a.shape[0] * 2, a.shape[1]), dtype=a.dtype)
    c[0::2] = a
    c[1::2] = b

    np.savetxt(
        results_file, np.vstack((mjd_time, roll_ang, c)).T, delimiter=" ", header=header
    )


if __name__ == "__main__":
    pass




from matplotlib import pyplot as plt
import numpy as np

from pyarchi.star_track import star_tracking_handler

from pyarchi.utils import my_timer
from pyarchi.utils import create_logger

logger = create_logger("main")


@my_timer
def photometry(data_fits, **kwargs):
    """
    This function ensures that the entire process runs in the correct order. First the masks are updated and, afterwards,
    the centers for the next image are calculated.

    Parameters
    ------------------
    data_fits: Object of "Data" class that will hold the results of the analysis for each star

    kwargs:
        dictionary with all the necessary information. It is loaded from the config.yaml file

    Returns
    ------------------
        data_fits: Object of "Data" class that holds the results of the analysis for each star
        -1: If an error was found during the process
    """

    if not kwargs[
        "optimize"
    ]:  # avoid excessive writing to log file during optimization process
        logger.info(
            "Starting data analysis process with a grid of {}".format(kwargs["grid_bg"])
        )

    if kwargs["headless"]:
        plt.switch_backend("Agg")
    else:
        plt.switch_backend("TkAgg")

    for index in range(data_fits.image_number):
        data_fits.update_stars(index)
        data_fits.calculate_uncertainties(index)

        if data_fits.abort_process:
            logger.fatal("Errors found during run time")
            return -1

        if kwargs["plot_realtime"] and not kwargs["optimize"]:
            pnt_mask = np.zeros(data_fits.stars[0].latest_mask.shape)
            for star in data_fits.stars:
                plt.contour(star.latest_mask)

                positions = star.positions[-1]

                pnt_mask[int(positions[0]), int(positions[1])] = 3e7

            img = data_fits.get_image(index).copy()
            plt.imshow(img)
            plt.contour(pnt_mask)

            plt.pause(0.2)

            plt.clf()

        # Update the star's positions for the next frame ##############

        points = star_tracking_handler(data_fits, index, **kwargs)

        if points == -1:
            logger.fatal("Errors found during center determination")
            return -1

    return data_fits



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyarchi',
 'pyarchi.data_objects',
 'pyarchi.initial_detection',
 'pyarchi.masks_creation',
 'pyarchi.output_creation',
 'pyarchi.routines',
 'pyarchi.star_track',
 'pyarchi.utils',
 'pyarchi.utils.data_export',
 'pyarchi.utils.factors_handler',
 'pyarchi.utils.misc',
 'pyarchi.utils.noise_metrics',
 'pyarchi.utils.optimization']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=4.0,<5.0',
 'matplotlib>=3.1.2,<4.0.0',
 'opencv-python>=4.1.2,<5.0.0',
 'pyyaml>=5.3,<6.0',
 'scipy>=1.4.1,<2.0.0']

setup_kwargs = {
    'name': 'pyarchi',
    'version': '1.0.0',
    'description': "Photometry for CHEOPS's background stars",
    'long_description': None,
    'author': 'Kamuish',
    'author_email': 'andremiguel952@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from .misc.logger_setup import create_logger

from .optimization.threaded_optimization import general_optimizer
from .factors_handler.factors_handlers import store_optimized_radius, get_optimized_mask


from .misc.timer import my_timer
from .misc.path_searcher import path_finder

from .misc.parameters_validator import parameters_validator
from .misc.calculate_moments import calculate_moments

from .noise_metrics.CDPP import CDPP
from .noise_metrics.DRP_CDPP import DRP_CDPP

from .data_export.export_fits import create_fits
from .data_export.export_txt import export_txt
from .data_export.export_photo_info import photo_SaveInfo



import numpy as np
import json
import os

from pyarchi.utils import create_logger

logger = create_logger("util")

def get_file_name(path, storage_location):
    """
    Extracts the last part of the path, i.e. the dataset folder name, to store the optimized masks

    Example
    ---------
        >>> get_file_name("kamuish/teste")  
        >>> "teste"

        >>> get_file_name("kamuish/teste/")  
        >>> "teste"
    Parameters
    ----------
    path : string
        path to the base folder
    """

    
    splitted_path = path.split("/")
    file_name = splitted_path[-2] if not splitted_path[-1] else splitted_path[-1]
    json_full_path = os.path.join(storage_location, "{}.json".format(file_name))

    return json_full_path

def store_optimized_radius(new_dict, **kwargs):  # TODO: improve save for method combination
    """
    Updates the json file in which the optimization info is stored

    :param new_dict: dict wth the format { star index : radius }
    :return: nothing
    """

    logger.info("Storing optimized radius in the json file")

    json_path = kwargs["optimized_factors"]
    method = kwargs["method"]
    detect_mode = kwargs["detect_mode"]
    size_bg_grid = str(kwargs["grid_bg"])
    initial_method = kwargs["initial_detect"]

    json_full_path = get_file_name(kwargs["base_folder"], kwargs["optimized_factors"])

    print(json_full_path)
    if os.path.exists(json_full_path):
        with open(json_full_path, "r") as data_file:
            try:
                data_loaded = json.load(data_file)
            except Exception as e:
                logger.critical(
                    "Problem opening the json file with the optimized radius. Could not save data to file",
                    exc_info=True,
                )
                return -1

        with open(json_full_path, "w") as data_file:
            try:
                for key, value in data_loaded[detect_mode][method][
                    size_bg_grid
                ].items():  # saving over old values
                    if not np.isnan(new_dict[key]):
                        data_loaded[detect_mode][method][size_bg_grid][initial_method][
                            key
                        ] = new_dict[key]
            except KeyError as e:
                try:  # initial_method does not exist
                    data_loaded[detect_mode][method][size_bg_grid][
                        initial_method
                    ] = new_dict

                except KeyError as e:  # bg_grid does not exist
                    try:
                        data_loaded[detect_mode][method][size_bg_grid] = {
                            initial_method: new_dict
                        }
                    except KeyError as e:
                        try:
                            data_loaded[detect_mode][method] = {
                                size_bg_grid: {initial_method: new_dict}
                            }
                        except KeyError as e:
                            data_loaded[detect_mode] = {
                                method: {size_bg_grid: {initial_method: new_dict}}
                            }

            json.dump(data_loaded, data_file, indent=4)

    else:
        logger.info("Json file with optimized radii not found. Creating new file")

        with open(json_full_path, "w") as data_file:
            json.dump(
                {detect_mode: {method: {size_bg_grid: {initial_method: new_dict}}}},
                data_file,
                indent=4,
            )

    return 0


def get_optimized_mask(**kwargs): # TODO: improve load for method combination
    """
    Retrieves the optimized radius from the json file. From this file we load a dict in which each key is
    the method name and the associated value is a dict with information about each star

    Parameters
    ----------
    kwargs
        - method : used to retrieve the optimization parameters from the json file

    Returns
    -------
    data_loaded:
        dict in which the keys are the index of each star and the values are the radius for it. If the method name does not exist then it's returned a empty dict
    """

    logger.info("Retrieving optimized factors")

    json_path = kwargs["optimized_factors"]
    json_full_path = get_file_name(kwargs["base_folder"],kwargs["optimized_factors"])

    if os.path.exists(json_full_path):
        with open(json_full_path, "r") as data_file:
            try:
                data_loaded = json.load(data_file)
            except Exception as e:
                logger.critical(
                    "Json file with optimized radius not found", exc_info=True
                )
                return -1
            else:

                try:
                    return data_loaded[kwargs["detect_mode"]][kwargs["method"]][
                        str(kwargs["grid_bg"])
                    ][kwargs["initial_detect"]]
                except KeyError as e:
                    logger.critical(
                        "Method {} with grid {} and detect mode not found in the json file \n "
                        "You should run the optimization process before analysing the star."
                        "".format(
                            kwargs["method"], kwargs["grid_bg"], kwargs["detect_mode"]
                        ),
                        exc_info=True,
                    )

                    return -1

    else:
        with open(json_full_path, "w") as data_file:
            logger.fatal("JSON file did not exist. created an empty one")
            return -1


if __name__ == "__main__":
    import yaml

    with open(
        "/home/andre/PycharmProjects/CHEOPS_mult_dtct/configs/config.yaml", "r"
    ) as stream:
        kwargs = yaml.load(stream)

    x = get_optimized_mask(**kwargs)




import numpy as np


def rotate_points(Data_fits, img_number, to_calculate):
    if img_number + 1 < len(Data_fits.roll_ang):
        rot_mat = Data_fits.get_rot_mat(
            ((Data_fits.roll_ang[img_number + 1] - Data_fits.roll_ang[0]) * np.pi / 180),
            clockwise=True,
        )

        scaling_factor = Data_fits.bg_grid / 200 if Data_fits.bg_grid != 0 else 1

        delta_pos = [100, 100]
        delta_pos = np.multiply(delta_pos, scaling_factor) + np.floor(scaling_factor / 2)

        for index, star in enumerate(Data_fits.stars):
            if index not in to_calculate:
                continue

            pair = star.positions[0].copy()
            coords = np.dot(rot_mat, [[pair[0] - delta_pos[0]], [pair[1] - delta_pos[1]]])

            x = coords[0][0] + delta_pos[0]
            y = coords[1][0] + delta_pos[1]

            if star.number in to_calculate:
                star.positions.append([x, y])


def offsets_method(Data_fits, index, primary, secondary):
    """
    This function expands the functionality of rotate_points. After rotating the points we calculate the offset
    experienced by the central star, by calculating the deviation between the center location from the last two images.
    The offset center for the central star is simply obtained from the fits files, without rotating the previous point.
    
    Parameters
    ---------------
    Data_fits
         :class:`pyarchi.main.initial_loads.Data` object.
    index
        index of the image
    primary:
        Methodology to apply to the central star. If it's offsets then the central star is tracked using this method
    secondary:
        Methodology to apply to the outer stars. If it's offsets then they are tracked using this method

    Returns
    -------

    """

    if primary != "offsets" and secondary != "offsets":
        return

    to_calculate = []

    if secondary == "offsets":
        to_calculate = to_calculate + [star.number for star in Data_fits.stars[1:]]

    scaling_factor = Data_fits.bg_grid / 200 if Data_fits.bg_grid != 0 else 1

    if index + 1 < len(Data_fits.roll_ang):

        offsets = Data_fits.offsets[index + 1]
        off_y = -512 + offsets[0]
        off_x = -512 + offsets[1]

        if primary == "offsets":
            central = [off_x + 100, off_y + 100]
            central = np.multiply(central, scaling_factor) + np.floor(scaling_factor / 2)
            Data_fits.stars[0].add_center(central)
            positions_main = Data_fits.stars[0].positions

        else:
            positions_main = Data_fits.offsets

        center_shift = [
            positions_main[index + 1][0] - positions_main[0][0],
            positions_main[index + 1][1] - positions_main[0][1],
        ]

        rotate_points(Data_fits, index, to_calculate)

        for j in range(1, len(Data_fits.stars)):
            if j in to_calculate:
                Data_fits.stars[j].change_center(
                    -1,
                    [
                        Data_fits.stars[j].positions[-1][0] + center_shift[0],
                        Data_fits.stars[j].positions[-1][1] - center_shift[1],
                    ],
                )
    return 0



import yaml
from functools import wraps

from .photometric_process import photometry
from pyarchi.data_objects import Data

from pyarchi.utils import general_optimizer, store_optimized_radius
from pyarchi.utils import create_logger, parameters_validator

logger = create_logger("Photo Controller")


class Photo_controller:
    """
        Controller class that allows an easy interface with all the important functions in the module. If the no_optim parameter is set to False,
        then it automatically starts the optimization routine, at instantiation time of the class.
    """

    def __init__(self, job_number, config_path="configuration_files/config.yaml", no_optim=False):
        self.job_number = job_number
        logger.info("Loading Parameters:")

        if isinstance(config_path, str):
            with open(config_path, "r") as stream:
                kwargs = yaml.load(stream)
        else:  # we can pass a dict
            kwargs = config_path

        self.kwargs = kwargs
        self.data_fits = Data(kwargs['base_folder']) 

        if no_optim:
            self.kwargs["optimize"] = 0

        if self.kwargs[
            "optimize"
        ]:  # Finds the best factors to minimize Cv and updates the json file
            self.__optimize()

        self._completed_run = False

    def _check_parameters(func):  # pylint: disable=no-self-argument
        """
        Used to validate the error flag before running a function.
        Decorator used inside this class

        Returns
        -------

        """
        @wraps(func)
        def on_call(self, *args, **kwargs):

            if (
                func.__name__ == "run" and self.kwargs["optimize"]
            ):  # pylint: disable=no-member
                pass   # during optimization process there are no checks
            else:
                wrong_params, warnings, kwargs = parameters_validator(**self.kwargs)
                if any(warnings):
                    logger.fatal("\t Found Warnings: {}".format(len(warnings)))
                    [logger.fatal("\t \t" + warn) for warn in warnings]

                if any(wrong_params):
                    logger.fatal(
                        "Problem with the passed parameters. Process is going to stop now"
                    )
                    logger.fatal("\t Bad parameters: {}".format(len(wrong_params)))
                    [logger.fatal("\t \t" + wrong_param) for wrong_param in wrong_params]
                    return -1

            if func.__name__ == "__optimize": # pylint: disable=no-member
                return func(self)  # pylint: disable=not-callable
            else:
                return func(self, *args, **kwargs) # pylint: disable=not-callable

        return on_call

    @_check_parameters
    def run(self, DataFits = None, factor=None, **kwargs):
        """
        Calls the main method to run the star analysis pipeline

        Parameters
        ----------
        factor
            Used for the optimization process. During normal functioning process then so value should be passed
        kwargs
            Config values, retrieved from the file on the class initialization
        Returns
        -------
            data_fits: instance of :class:`~pyarchi.data_objects.Data.Data`, with the relevant information.
        """

        configs = (
            self.kwargs if (factor is None and not self.kwargs["optimize"]) else kwargs
        )

        data_fits = DataFits if DataFits is not None else self.data_fits
        result = data_fits.load_parameters(factor, **configs)

        if result == -1:
            logger.fatal("Critical error")
            return data_fits

        self.data_fits = photometry(data_fits = data_fits, **configs)

        if not self.kwargs["optimize"]:
            logger.info("Checking for out of bounds masks:")
            found = False 
            for star in self.data_fits.stars:
                if star.out_bound:
                    logger.warning("\t\t Star {} is out of bounds".format(star.number))
                    found = True 
            print("\t{}\n==============//==============".format("-> Found nothing" if not found else ''))
        if data_fits.abort_process:
            logger.fatal("Problems were found. Could not run properly !!!!!")
            raise Exception("Something went wrong")

        self._completed_run = True

        return self.data_fits

    def change_parameters(self, params_dict):
        """
        Updates the kwargs with new values. A dictionary is passed in, with keys corresponding to parameters and each
        value it's the new configuration for that specific parameter.
        At this stage no validation is made over the passed values. Such validation occurs before the routines, to make sure that the correct parameters were added.

        Parameters
        ----------
        params_dict:
            Dictionary with the values that we wish to change

        Returns
        -------

        """
        new_dict = self.kwargs.copy()
        for key, value in params_dict.items():
            if key not in self.kwargs:
                raise AttributeError(
                    "The {} configuration value does not exist".format(key)
                )
            if key == 'optimize' and value != 0:
                raise KeyError("Cannot set optimize parameter to one. Use the .optimize() method to manually trigger the optimization routine")
            new_dict[key] = value
        
        self.kwargs = new_dict
        self.data_fits.used_file = self.kwargs["base_folder"]

    def optimize(self):
        """
        Run the optimization routine without instantiating a new object.
        Returns
        -------

        """
        self.kwargs["optimize"] = 1  # just making sure
        self.__optimize()

    @_check_parameters
    def __optimize(self):
        """
        Optimization process. Called when the class is initialized
        Returns
        -------

        """
        import time

        t0 = time.time()
        logger.info("Starting optimization process")

        vals = general_optimizer(
            self.run,
            data_f=Data('---'),
            job_number=self.job_number,
            max_process=self.kwargs["optim_processes"],
            **self.kwargs
        )

        logger.info("Optimization process completed after {} s".format(time.time() - t0))
        self.kwargs["optimize"] = 0

        if store_optimized_radius(vals, **self.kwargs) == -1:
            logger.critical("Problems were found. Failed to save optimized values.")
            return
        del vals
        self._completed_run = False

    @property
    def parameters(self):
        return self.kwargs



from .dynamical_centers import dynam_method
from .offset_centers import offsets_method
from .static_method import static_method


def star_tracking_handler(Data_fits, index, **kwargs):
    """
    Handles the detection mode for the target star and the ones around it. Allows to have two different modes of
    detection active at the same time.

    Parameters
    ----------
    Data_fits:
         :class:`pyarchi.main.initial_loads.Data` object.
    index:
        Image number
    kwargs

    Returns
    -------
    int
        Error code of the different star tracking methods. 0 if everything as expected
        1 otherwise

    """

    if "+" not in kwargs["detect_mode"]:
        primary = secondary = kwargs["detect_mode"]
    else:
        primary, secondary = kwargs["detect_mode"].split("+")

    results = static_method(Data_fits, index, primary, secondary)

    if results == -1:
        return -1

    results = dynam_method(Data_fits, index, primary, secondary)

    if results == -1:
        return -1

    results = offsets_method(Data_fits, index, primary, secondary)
    return results



import os


def handle_folders(number_stars, job_number, singular=None, **kwargs):
    """
        If the folder structure is not present, then it is created. If we only wish to output data for a singular star, then no more folder are created.

    Parameters
    ----------
    number_stars:
        number of stars that will have a folder
    job_number:
            Job number, from the supernova server. Will be used to save the parent folder inside the
            kwargs["results_folder"] to create the entire directory structure.
    singular:
        If it's not None, then it should be a number. Only has an effect over the graphs but, on the files, the
        light curves are all saved.
    kwargs

    Returns
    -------

    """

    path = os.path.join(kwargs["results_folder"], "{}".format(job_number))

    if not os.path.exists(path):
        os.mkdir(path)

    for j in range(number_stars):

        if singular is not None and j != singular:
            continue

        dir_name = "Star_{}".format(j)

        star_path = os.path.join(path, dir_name)
        if not os.path.exists(star_path):
            os.mkdir(star_path)

    return path


if __name__ == "__main__":
    handle_folders(2, 1, **{"results_folder": "."})



import numpy as np


def change_grid(small_grid, increase_factor):
    """
    Changes from the small grid to a big grid. The increase is given by the increase factor.

    For example, if we have a [200,200] image and use an increase_factor of 3, an array with the shape [600,600]
    would be returned.
    Parameters
    ----------
    small_grid:
        Base array
    increase_factor:
        How much we should increase the array

    Returns
    -------

    """

    tiled = np.repeat(small_grid, increase_factor, axis=0)
    tiled = np.repeat(tiled, increase_factor, axis=1)

    return tiled



from astropy.io import fits
import numpy as np

from pyarchi.masks_creation import create_circular_mask, create_shape_mask
from pyarchi.masks_creation import change_grid

from pyarchi.initial_detection import centers_from_fits, initial_dynam_centers

from pyarchi.utils.misc.rotation_mats import matrix_cnter_clock, matrix_clockwise
from pyarchi.utils import path_finder
from pyarchi.utils import get_optimized_mask
from pyarchi.utils import create_logger

from .Star_class import Star

logger = create_logger("Data")


class Data:
    """
        Data structure to hold the information of all the stars in the field.

        Parameters
        ---------------
        base_folder:
            location of the folder in which all the fits files are located.
        roll_ang:
            rotation angle for all the images.
        imgs:
            all the images in the fits file.
        stars:
            List in which all elements are a "Star" object that holds information of one star.
        mask_type:
            type of mask. either circle or shape 
        detect_mode:
            detection mode used
        init_detection_mode:
            initial detection mode
        mjd_time:
            mjd time of the observations
    """

    def __init__(self, filename):
        # error flag
        self._error_flag = 0

        # configuration loaded from the .yaml
        self.base_folder = ""
        self.mask_type = ""
        self.detect_mode = ""
        self.init_detection_mode = ""
        self.used_file = filename
        self.bg_grid = 0

        # data loaded from the fits file
        self.roll_ang = None
        self.mjd_time = None
        self._imgs = None
        self.offsets = None
        self.image_number = None
        self.forbidden_regions = []

        # misc
        self._stars = []
        self._image_dict = {}
        self.low_mem = 0
        self.calc_uncert = False
        self.uncertainties_params = {}  # to estimate the uncertainties for our data

    def _reset(self):
        """
        Resets the object to its initial state. "Cleaning" all data previously stored in here.

        It's called when the parameters are loaded on :function:`pyarchi.main.initial_loads.Data.load_parameters`.
        Keep in mind that the reset routine does not re-enable the disabled stars.
        Returns
        -------

        """

        # error flag
        self._error_flag = 0

        # configuration loaded from the .yaml
        self.base_folder = ""
        self.mask_type = ""
        self.detect_mode = ""
        self.bg_grid = 0

        # data loaded from the fits file
        self.roll_ang = None
        self.mjd_time = None
        self._imgs = None
        self.offsets = None
        self.forbidden_regions = []

        self._image_dict = {}

        # misc
        for star in self._stars:
            star.remove_data()

        Star.number = 0  # Reset the number of the stars every time this routine is called

    def _verify_validity(func):  # pylint: disable=no-self-argument
        """
        Used to validate the error flag before running a function.
        Decorator used inside this class

        Returns
        -------

        """

        def on_call(self, *args, **kwargs):
            if self._error_flag:
                logger.fatal(
                    "Error found. Method {} is not going to run. ".format(
                        func.__name__
                    )  # pylint: disable=no-member
                )
                return -1
            else:
                return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return on_call

    def load_parameters(self, factor=None, **kwargs):
        """
        Loads all the necessary information from the .fits files. Launches the routines to find the initial
        position of the stars and create the corresponding masks.

        Before running the internal parameters, as well as the star's ones are removed. This was made in order to
        minimize loading data from disk. However, keep in mind that the reset routine does not re-enable stars. So, if a
        star was disabled in this object, then it will stay that way unless a new object is instantiated.
        Parameters
        ----------
        factor
        kwargs

        Returns
        -------
        -1:
            Errors were found during runtime
        0:
            Everything went without problems

        """
        self._reset()
        self.base_folder = kwargs["base_folder"]
        self.init_detection_mode = kwargs['initial_detect']
        self.detect_mode = kwargs["detect_mode"]
        self.mask_type = kwargs["method"]
        self.bg_grid = kwargs["grid_bg"]
        self.low_mem = kwargs["low_memory"]
        self.calc_uncert = kwargs["uncertainties"]

        load_flag = self._init_load(**kwargs)
        pos_flag = self._init_pos(**kwargs)
        mask_flag = self._initialize_masks(factor, **kwargs)

        if -1 in [load_flag, pos_flag, mask_flag]:
            self._error_flag = 1
            return -1
        else:
            init_image = self.get_image(0)

            self.forbidden_regions = np.where(np.isnan(init_image))
        return 0

    @_verify_validity
    def update_stars(self, image_number):
        """
        Controls the data treatment process for each image. Updates the masks for each star and calculates the flux
        passing through our mask. Afterwards the mask is validated to see if we have overlaps with the forbidden region,
        i.e., if the mask is over the  NaN areas of the image.
        Parameters
        ----------
        image_number

        Returns
        -------

        """

        scaling_factor = self.bg_grid / 200 if self.bg_grid else 1
        for star in self._stars:
            if not star.is_active:
                continue

            star.update_mask(scaling_factor, image_number)

            final_img = self.get_image(image_number)
            try:
                final_result = np.multiply(star.latest_mask, final_img)
            except Exception as e:
                print(image_number)
                print(star, star.latest_mask.shape, final_img.shape)
                raise e

            photom = np.nansum(final_result)

            star.photom.append(photom)

        self._validate_forbidden_region()

    def get_image(self, number):
        """
        Ask for a given image. If the background grid is in use, returns the increased image
        
        Parameters
        ----------
        number : int
            Image number
        
        Returns
        -------
        image: numpy array
            Desired image
        """
        if number == "all":
            return list(self._image_dict.values()) if self._image_dict else self._imgs

        if self.low_mem or self.bg_grid:
            self._increase_imgs(number)
            return self._image_dict[number].copy()
        else:
            return self._imgs[number].copy()

    def reload_images(self, **kwargs):
        """
        Reload images from disk, in the case that the desired image has been deleted
        Parameters
        ----------
        kwargs

        Returns
        -------

        """
        return self._init_load(**kwargs)

    def _init_load(self, **kwargs):
        """
                Opens the fits files and extracts the roll angle and the images

        Parameters
        ----------
        kwargs

        Returns
        -------

        """

        self._image_dict = {}

        logger.info("Extracting official pipeline Lightcurve information")

        default_path = path_finder(mode="default", **kwargs)
        subarray_path = path_finder(mode="subarray", **kwargs)
        try:
            hdulist = fits.open(default_path)
        except IOError:
            logger.error("Official Lightcurve file not found")
            self._error_flag = 1
            return -1
        else:
            with hdulist:
                self.roll_ang = hdulist[1].data["ROLL_ANGLE"]
                self.mjd_time = hdulist[1].data["MJD_TIME"]
                self.offsets = list(
                    zip(hdulist[1].data["CENTROID_X"], hdulist[1].data["CENTROID_Y"])
                )

                if self.calc_uncert:
                    possible_curves = ["DEFAULT", "OPTIMAL", "RSUP", "RINF"]
                    possible_curves.remove(kwargs["official_curve"])
                    darks = []

                    def_points = np.pi * (hdulist[1].header["AP_RADI"]) ** 2
                    self.uncertainties_params["bg"] = (
                        hdulist[1].data["BACKGROUND"] / def_points
                    )
                    curr_dark = hdulist[1].data["DARK"] / def_points
                    darks.append(curr_dark)

                    for curve in possible_curves:
                        default_path = path_finder(
                            mode="default", off_curve=curve, **kwargs
                        )
                        with fits.open(default_path) as file:
                            def_points = (
                                np.pi * (file[1].header["AP_RADI"]) ** 2
                            )  # pylint: disable=no-member
                            curr_dark = (
                                file[1].data["DARK"] / def_points
                            )  # pylint: disable=no-member
                            darks.append(curr_dark)

                    self.uncertainties_params["dark"] = np.median(darks, axis=0)
                    self.uncertainties_params["t_exp"] = hdulist[1].header["EXPTIME"]
                    self.uncertainties_params["nstack"] = hdulist[1].header["NEXP"]
        try:
            hdulist = fits.open(subarray_path)
        except IOError:
            logger.error("Subarray file not found")
            self._error_flag = 1
            return -1
        else:
            with hdulist:
                imgs = hdulist[1].data
                self._imgs = imgs.astype(np.float)
                self.image_number = self._imgs.shape[0]

                if self.calc_uncert:
                    self.uncertainties_params["cron"] = 1.96 * hdulist[2].data["RON"]

    @_verify_validity
    def _init_pos(self, **kwargs):
        """
        Extracts the initial position of each star, creating a Star object for each of  those positions and
        storing them inside the self._stars. We can have the initial position determination with an image treatment
        approach (dynam), by loading data from the fits files and inferring their position (fits) or, specify a load
        type for the central star and the outer ones, separating them both with a plus sign..

        :return:
        """

        if "+" not in kwargs["initial_detect"]:
            primary = secondary = kwargs["initial_detect"]
        else:
            primary, secondary = kwargs["initial_detect"].split("+")

        disabled = [i.number for i in self.stars if not i.is_active]

        self._stars = initial_dynam_centers(self.get_image(0), self.bg_grid, **kwargs)
        
        for j, star in enumerate(self._stars):
            if j in disabled:
                star.disable()

        self._stars = centers_from_fits(
            primary, secondary, self._stars, self.roll_ang[0], self.offsets[0], **kwargs
        )


    @_verify_validity
    def _initialize_masks(self, factor, **kwargs):
        """
        Controls the process of creating the masks for each star. During the optimization process the factor is
        passed as an argument. Otherwise, it is loaded from the corresponding file.

        Parameters
        ----------
        factor:
            Passed during the optimization process. Mask is created using it it something other than None is passed
        kwargs

        Returns
        -------

        """

        scaling_factor = self.bg_grid / 200 if self.bg_grid else 1
        if factor is None:
            # If we have no manual input and no  command for optimization
            # then we will load optimized values from the json file

            optim_mask = get_optimized_mask(**kwargs)
            if optim_mask == -1:
                logger.fatal("Could not retrieve the optimized factors")
                self._error_flag = 1
                return -1

            if self._create_mask(optim_mask, scaling_factor, **kwargs) == -1:
                logger.fatal("Problem with the mask creation process")
                return -1

        else:  # during optimization process
            self._create_mask(factor, scaling_factor, **kwargs)

    @_verify_validity
    def _create_mask(self, factor, scaling_factor, **kwargs):
        """
        Creates a mask for each star, using the specified method on the config.yaml file

        Parameters
        ----------
        factor:
            Scaling factor - integer if this is called during the optimizaton process. Otherwise should be a list

        size_grid_change:
            size of the big grid


        Returns
        -------

        """

        if self.abort_process:
            logger.fatal("Process was aborted due to previous errors")
            return -1

        if "+" not in kwargs["method"]:
            primary = secondary = kwargs["method"]
        else:
            primary, secondary = kwargs["method"].split("+")

        circle_result = create_circular_mask(
            self.get_image(0), self._stars, factor, primary, secondary
        )
        if circle_result == -1:
            logger.fatal("Could not create mask using the 'CIRCLE' method. ")
            return -1
        shape_result = create_shape_mask(
            self.get_image(0), self._stars, factor, scaling_factor, primary, secondary
        )

        if shape_result == -1:
            logger.fatal("Could not create mask using the 'SHAPE' method. ")
            return -1

        # Create a joint dict with the star number and the mask
        full_dict = circle_result.copy()
        full_dict.update(shape_result)

        for key, mask in full_dict.items():
            process_flag = self._stars[key].add_initial_mask(
                mask, factor, scaling_factor, kwargs["low_memory"]
            )

            if process_flag == -1:
                return -1

        return 0

    def calculate_uncertainties(self, index):
        """
        Trigger the calculation of the uncertainties
        
        Parameters
        ----------
        index : int
            Image number
        """
        if not self.calc_uncert:
            return
        else:
            params = self.uncertainties_params
            for star in self.stars:
                star.calculate_errors(
                    background=params["bg"][index],
                    nstack=params["nstack"],
                    cron=params["cron"][index],
                    dark=params["dark"][index],
                )

    def _validate_forbidden_region(self):
        """
        Checks for overlaps with the forbidden region
        Returns
        -------

        """

        for star in self._stars:

            if star.out_bound or not star.is_active:
                continue

            if star.latest_mask[self.forbidden_regions].any() != 0:
                star.out_bound = True

    def disable_star(self, star_number):
        """
        Disables a star to avoid computing redundant information i.e. after the best value has been found during the
        optimization process for one star, but others are still not yet optimized.
        Parameters
        ----------
        star_number

        Returns
        -------

        """
        self._stars[star_number].disable()

    @_verify_validity
    def _increase_imgs(self, index):
        """
        Increases the size of all images before the process starts
        Returns
        -------

        """

        if index not in self._image_dict:
            ratio = self.bg_grid / 200
            if ratio == 0:
                im = self._imgs[0].copy()
            else:
                im = np.divide(change_grid(self._imgs[0].copy(), ratio), ratio ** 2)

            self._imgs = self._imgs[1:]

            self._image_dict[index] = im

            if self.low_mem:
                self._image_dict = {
                    key: val for key, val in self._image_dict.items() if key > index - 2
                }

    # Properties of the class
    @property
    def stars(self):
        """
        Returns the list with all of the :class:`~pyarchi.data_objects.Star_class.Star` objects.
        If an error has been raised, returns an empty list
        
        """
        if not self._error_flag:
            return self._stars
        else:
            return []

    @property
    def abort_process(self):
        """
        If the error_flag is set then the process should be aborted
        Returns
        -------

        """
        return self._error_flag

    @property
    def all_curves(self):
        """
        Light curve of all stars, in the star numbering order (from closest to center to farthest)
        Returns
        -------

        """
        if not self._error_flag:
            return np.asarray([star.photom for star in self.stars])
        else:
            return []

    @property
    def all_uncertainties(self):
        """
        Light curve of all stars, in the star numbering order (from closest to center to farthest)
        Returns
        -------

        """
        if not self._error_flag:
            return np.asarray([star.uncertainties for star in self.stars])
        else:
            return []

    @property
    def is_empty(self):
        """
        See if the object has been used in a run, i.e. if it has data
        
        Returns
        -------
        boolean:
            True if it is empty or an error has been raised, False otherwise
        """
        if not self.stars:
            return True
        return len(self.stars[0].photom) == 0 or self._error_flag

    # Static methods
    @staticmethod
    def get_rot_mat(angle, clockwise):
        """
        Returns a rotation matrix (around the origin) for a given angle
        :param angle: rotation angle, in degrees
        :param clockwise: True for a clockwise rotation and False for a counter clockwise rotation.
        :return: rotation matrix
        """
        if clockwise:
            return matrix_clockwise(angle)
        else:
            return matrix_cnter_clock(angle)



import numpy as np
import os
import logging

from .optimizer import optimizer
from .circular_fine_tune import circular_tuner

from pyarchi.utils import create_logger

logger = create_logger("utils")


def general_optimizer(func, data_f, job_number, max_process, **kwargs):
    """
    Responsible for setting up the variables used during the optimization process.

    If the factor determined to be the best one is near the upper limit, then the range of values is extended and a new
    search for minimum noise starts. If the determined factor is inside a "safe" distance away from the highest possible
    value then the star is disabled and no further studies are made on it.

    Also responsible for creating .txt files with all the relevant information
    Parameters
    ----------
    func:
        function that launches the photometric process
    data_f:
        :class:`pyarchi.main.initial_loads.Data`  object with all the stars information inside
    job_number:
        JOb number assigned by the SLURM workload manager
    max_process:
        Maximum number of processes that can be launched during this routine
    Returns
    -------
    optimized_dict:
        Dictionary in which the keys are the star numbers and the values are the optimal factors
    """
    logger.info("Optimization with grid of: {}".format(kwargs["grid_bg"]))

    low, high = kwargs["val_range"]
    step = kwargs["step"]
    logging.disable(logging.INFO)

    value_range = np.arange(low, high + step, step)
    path = os.path.join(kwargs["results_folder"], str(job_number))
    if not os.path.exists(path):
        os.mkdir(path)

    file_path = os.path.join(path, "optimization_info.txt")

    with open(file_path, mode="w") as file:
        file.write("Method - {}\n".format(kwargs["method"]))
        file.write("Detect mode - {}\n".format(kwargs["detect_mode"]))
        file.write("Initial detection mode - {}\n".format(kwargs["initial_detect"]))
        file.write("Background grid - {}\n".format(kwargs["grid_bg"]))
        file.write("Factors Stars: -> \n")

    min_cvs, optimized_dict = optimizer(
        value_range, max_process, func, data_f, file_path, **kwargs
    )

    increase = high - low
    iterations = 0
    to_disable = []

    while len(to_disable) != len(optimized_dict.values()):
        to_disable = []

        for key, val in optimized_dict.items():
            if val < high - 2 * step:  # allow for some tolerance near the border
                to_disable.append(int(key))

        print(
            "DISABLING: {} - Last run in the interval: {} <-> {}".format(
                to_disable, low, high
            )
        )
        if len(to_disable) == len(optimized_dict):
            logger.fatal("All stars were disabled")
            break

        low, high = high - 2 * step, high + increase - 2 * step
        value_range = np.arange(low, high + step, step)
        min_cvs_tmp, optimized_dict_tmp = optimizer(
            value_range, max_process, func, data_f, file_path, to_disable, **kwargs
        )

        for (
            star,
            cdpp,
        ) in min_cvs_tmp.items():  # update old values if the new ones are better
            if cdpp < min_cvs[star]:
                min_cvs[star] = cdpp
                optimized_dict[str(star)] = optimized_dict_tmp[str(star)]

        iterations += 1
        if iterations > kwargs["optimization_extensions"]:
            logger.fatal("Reached max iters. Last values: {} {}".format(low, high))
            break

    if kwargs["fine_tune_circle"] and "circle" in kwargs["method"]:
        vals = kwargs["method"].split("+")

        to_disable = []

        if len(vals) == 1 and vals[0] == "circle":
            to_disable = []

        elif vals[0] == "circle":
            to_disable = [i for i in data_f.stars[1:].number]
        elif len(vals) != 1 and vals[1] == "circle":
            to_disable = [0]

        min_cvs_tmp, optimized_dict_tmp = circular_tuner(
            optimized_dict.values(),
            max_process,
            func,
            data_f,
            file_path,
            to_disable,
            **kwargs
        )

        for (
            star,
            cdpp,
        ) in min_cvs_tmp.items():  # update old values if the new ones are better
            if cdpp < min_cvs[star]:
                min_cvs[star] = cdpp
                optimized_dict[str(star)] = optimized_dict_tmp[str(star)]

    logging.disable(logging.NOTSET)

    logger.info("Max factor used: {}".format(high))
    [
        logger.info(
            "Star {} -  Factor: {} - Noise: {} ppm".format(
                index, optimized_dict[index], min_cvs[int(index)]
            )
        )
        for index in optimized_dict
    ]

    return optimized_dict
